#include <cstring> /* for memset */
#include <cstdio>  /* for printf,fopen,fread */
#include <cmath>   /* for cos,exp */
#include <cassert> /* for assert */
#include <algorithm> /* for remove_if */
#include <signal.h>

/* S3M player (plays both PCM and AdLib samples)
 * Also converts the song to MIDI.
 *
 * Written by and copyright (C) 2012 Joel Yliluoma - http://iki.fi/bisqwit/
 *
 * To produce MIDI, you will need to customize the instr_specs, per each song.
 *
 * You may also want to tune GlobalFineTuning.
 * And MaxMIDIchannel and BendSense (optional), and possibly SamplingRate.
 * Sorry that those are currently not commandline options.
 *
 * Use MaxMIDIchannel = 16 to produce MIDI files that can be played back
 * on those MIDI players that do not support the multi-synthesizer option.
 * It will cause channel congestion, but it is unavoidable if multi-synthesizer
 * option (FF 09 event, "choose device") is not available.
 *
 */

#include <map>
#include <vector>

#define HAS_YMF262 1
#if HAS_YMF262
# include "ymf262.c"  /* Use -DINLINE=inline '-Dlogerror(...)'   for compile */
#endif

typedef int int32_t;

#include "midi.hh"

static const bool DoLanczosFiltering = false;

static const unsigned IRQRate      = 200u;
static const unsigned SamplingRate = 44100u;
static const unsigned MaxPCMsamplesPerIRQ = (SamplingRate+IRQRate-1) / IRQRate;

namespace MIDItracer
{
    struct Instr_spec
    {
        struct Part
        {
            int patch;
            double tuning; // +- seminotes (12 = octave)
            double volume; // 1.0 = unmodified
        };
        std::map<int/*sample offset*/, std::vector<Part> > formula;
    };

    // Instrument mapping for FALL3.S3M
    static const std::vector<Instr_spec> instr_specs =
    {
        // 0: Nothing (never anything)
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{} },
#if 1 /* FALL3.S3M */
        // 1: Minor synstring chord
        // C C# D D# E F F# G
        // 0 1  2 3  4 5 6  7
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {68,0-0-0.1   ,0.9 * 0.62},
                          {68,3-0-0.1   ,0.9 * 0.62},
                          {68,7-0-0.1   ,0.9 * 0.62},
                          {84,0-24+0.1   ,0.7 * 0.62},
                          {84,3-24+0.1   ,0.7 * 0.62},
                          {84,7-24+0.1   ,0.7 * 0.62},
                          {50,0-12  ,0.6 * 0.62},
                          {50,3-12  ,0.6 * 0.62},
                          {50,7-12  ,0.6 * 0.62}, } } }},
        // 2: Major synstring chord
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {68,0-0-0.1   ,0.9 * 0.62},
                          {68,4-0-0.1   ,0.9 * 0.62},
                          {68,7-0-0.1   ,0.9 * 0.62},
                          {84,0-24+0.1   ,0.7 * 0.62},
                          {84,4-24+0.1   ,0.7 * 0.62},
                          {84,7-24+0.1   ,0.7 * 0.62},
                          {50,0-12  ,0.6 * 0.62},
                          {50,4-12  ,0.6 * 0.62},
                          {50,7-12  ,0.6 * 0.62}, } } }},
        // 3: Bright pluck (atmosphere?)
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {104, 24, 1.8 * 0.62},
                          {100, 12, 1.0 * 0.62} } } }},
        // 4: Nothing
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{} },
        // 5: Massive synth brass
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { { 63,  0,2.1 * 0.62},
                          { 63,-12,1.6 * 0.62}, } } }},
        // 6: Synth contrabass pluck or something like that
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { { 39,-12-12,1.1 * 0.62},
                          { 39,-12-24,1.1 * 0.62}, } } }},
        // 7: Synth contrabass pluck or something like that: Little bit softer
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { { 38,-12-12,1.2 * 0.62}, } } }},
        // 8: Synth bass, short, octave higher than 6 and 7
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { { 39,-12- 0,1.2 * 0.62}, } } }},
        // 9: Synth bass, short; little more pressuring
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { { 38,-12 -0,1.2 * 0.62},
                          { 38,-12-12,1.2 * 0.62}, } } }},
        // 10: Nothing
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{} },
        // 11: Reverse cymbal
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {120,  0, 1.0 * 3}, } } }},
        // 12: Bass drum + soft hihat a bit later
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {117, -12,1.0 * 63},
                          //{117,   6,1.0 * 3},
                          {117, -24,1.0 * 3},
                          {-35,   0,1.0 * 43},
                          {-36,   0,1.0 * 43} }/*,
             }, { 2170, { {-46, 0,  1.0 * 0.2}, }*/ } }},
        // 13: Massive snare
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {-40,  0, 1.0 * 1.4},
                          {117,-24, 1.0 * 1.4},
                          {-36,   0,1.0 * 3} } } }},
        // 14: Synth brass lead: The main melodic instrument
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { { 87,  0, 3.5 * 0.62},        // Fifth
                          { 50, 12, 1.7 * 0.62} } } }}, // Strings2
        // 15: Open high hihat: tsihhh
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {-46,  0, 1.0 * 4.0},
                          {-54,  0, 1.0 * 2.5}, } } }},
        // 16: Closed high hihat: zi
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {-42,  0, 1.0 * 2.4}, } } }},
#endif
#if 0 /* H-SPACE */
        // 1: Slow attack string
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {63,12    ,1.1},
                          {48, 0    ,0.7},
                          {50,-12   ,1.3},
                        } } }},
        // 2: Harp pluck
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {47,7     ,1.3},
                          {100,7+12, 0.6} } } }},
        // 3: Bass/brass no attack, cuts abruptly
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {63,-12   ,1.8},
                          {36,  0,   1.0} } } }},
        // 4: Tubular bells?
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {15,12   ,1.7},
                          {14, 0   ,0.5}, } } }},
        // 5: Massive snare
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {-40,  0, 1.0 * 1.4},
                          {117,-24, 1.0 * 1.4}, } } }},
        // 6: Bass drum
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {117, -12,1.0 * 3},
                          {-35,   0,1.0 * 3} } } }},
        // 7: Closed high hihat
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {-42,  0, 1.0 * 1.4}, } } }},
#endif
#if 0 /* 8bitbubsy fox */
        // 1: Alto sax?
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {66, 0   ,2.1},
                        } } }},
        // 2: Mute trumpet?
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {60,-24  ,1.6},
                        } } }},
        // 3: Alto sax?
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {66,12   ,2.1},
                        } } }},
        // 4: Clarinet
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {72,12   ,2.1},
                        } } }},
        // 5: Bas drum
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {-35, 0   ,5.1},
                        } } }},
        // 6: Hihat short
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {-44, 0   ,5.1},
                        } } }},
        // 7: Snare
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {-38, 0   ,3.1},
                        } } }},
        // 8: Mute trumpet, again
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { { 60,12   ,2.1},
                        } } }},
#endif
#if 0 /* satellite one */
        // 1: Power bass loop
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { { 63,-3.1-12+  0+12,2.6*1.3},
                          { 63,-3.1-12+-12+12,1.2*1.3}, } } }},
        // 2: Bass drum
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {117, -12,1.0 * 3},
                          {-35,   0,1.0 * 3},
                          {-36,   0,1.0 * 3} } } }},
        // 3: Klooh
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { { 18,   0,0.3*0.8},
                          { 13,   0,0.4*0.8},
                          {  3,   0,0.2*0.8},
                          { 59,   0,1.2*0.8},
                         } } }},
        // 4: Synth string? Sawwave
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { { 82,   0,1.3} } } }},
        // 5: Oboe
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { { 69,   0,8.0},
                          { 54, -12,1.0}} } }},
        // 6: Hihat mid
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {-49, 0   ,5.1},
                  } } }},
        // 7: Massive snare
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {-40,  0, 1.0 * 1.4},
                          {117,-24, 1.0 * 1.4},
                          {117, -12,1.0 * 3},
                          {-35,   0,1.0 * 3},
                          {-36,   0,1.0 * 3} } } }},
        // 8: Hihat short
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {-44, 0   ,5.1},
                        } } }},
#endif


#if 0 /* 4-mat's "That Old Magic" */
        // 1: Bass drum
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {117, -12,1.0 * 5},
                          {-35,   0,1.0 * 5},
                          {-36,   0,1.0 * 5} } } }},
        // 2: Massive snare
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {-40,  0, 1.0 * 1.4},
                          {117,-24, 1.0 * 1.4},
                          {117, -12,1.0 * 3},
                          {-35,   0,1.0 * 3},
                          {-36,   0,1.0 * 3} } } }},
        // 3: Synth bass (+10)
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { { 38,  -2,  0.7},
                          { 34, -14,  0.6},
                          { 30, -14,  0.4}, } } }},
        // 4: Strings loop D major  (D F# A >D) (-1, -6,-9,-13)
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {68,-1-0  +24 ,0.9*0.5},
                          {68,-6-0  +24 ,0.9*0.5},
                          {68,-9-0  +24 ,0.9*0.5},
                          {68,-13-0 +24 ,0.9*0.5},
                          {84,-1-24 +24 ,0.7*0.5},
                          {84,-6-24 +24 ,0.7*0.5},
                          {84,-9-24 +24 ,0.7*0.5},
                          {84,-13-24+24 ,0.7*0.5},
                          {50,-1-12 +24 ,0.6*0.5},
                          {50,-6-12 +24 ,0.6*0.5},
                          {50,-9-12 +24 ,0.6*0.5},
                          {50,-13-12+24 ,0.6*0.5}, } } }},
        // 5: Strings loop D major7 (D F# A B) (-4, -6, -9, -13)
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {68,-4-0  +24 ,0.9*0.5},
                          {68,-6-0  +24 ,0.9*0.5},
                          {68,-9-0  +24 ,0.9*0.5},
                          {68,-13-0 +24 ,0.9*0.5},
                          {84,-4-24 +24 ,0.7*0.5},
                          {84,-6-24 +24 ,0.7*0.5},
                          {84,-9-24 +24 ,0.7*0.5},
                          {84,-13-24+24 ,0.7*0.5},
                          {50,-4-12 +24 ,0.6*0.5},
                          {50,-6-12 +24 ,0.6*0.5},
                          {50,-9-12 +24 ,0.6*0.5},
                          {50,-13-12+24  ,0.6*0.5}, } } }},
        // 6: Closed high hihat
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {-42,  0, 1.0 * 6.4}, } } }},
        // 7: Open high hihat
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {-46,  0, 1.0 * 6.5}, } } }},
        // 8: Pan flute (>D, G,A) (14,7,9)
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {76, 14+0, 1.0 * 3.5} },
             }, { 1400, { {76,  7+0, 1.0 * 3.5} },
             }, { 2650, { {76,  9+0, 1.0 * 3.5} } } }},
        // 9: Distortion guitar thomp v1 (2466 samples) (tuned to <A = -3)
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { { 28, -3, 1.0 * 0.5},
                          { 28,-15, 1.0 * 0.5},
                          { 30,-15, 1.0 * 0.5} } } }},
        // 10: Distortion guitar thomp v2 (2194 samples)
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { { 28, -3, 1.0 * 0.5},
                          { 28,-15, 1.0 * 0.5},
                          { 28,  9, 1.0 * 0.5} } } }},
        // 11: Distortion guitar thomp v3
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { { 28, -3, 1.0 * 0.5},
                          { 28,-15, 1.0 * 0.5},
                          { 28,-27, 1.0 * 0.5} } } }},
        // 12: Distortion guitar thomp v4 (A + D) (-3, -10)
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { { 28, -3, 1.0 * 0.5},
                          { 28,-10, 1.0 * 0.5},
                          { 28,-15, 1.0 * 0.5} } } }},
        // 13: Trumpet or oboe, +1 semitone (tuned to A# = 10)
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { { 69, 10, 1.0 * 2.0},
                          { 60, 10, 1.0 * 1.2}
                        } } }},
        // 14: Vibrato'd strings, or ghoul or something
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { { 104, 10, 1.0 * 0.3},
                          { 101,-14, 1.0 * 0.3},
                          {  97,-14, 1.0 * 0.3},
                          { 102, 34, 1.0 * 0.3}} } }},
        // 15: Snare + car horn with fadeout (15, 13,10) (base note: C is same as #17 + F)
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { { -38,  0, 1.0 * 6.5},
                          {  77, 15, 1.0 * 0.4}, // bottle blow
                          {  56, 15, 1.0 * 0.4}, // orch hit
                          {  58, 15, 1.0 * 0.4}, // trump
                          {  58, 13, 1.0 * 0.4},
                          {  58, 10, 1.0 * 0.4}} } }},
        // 16: Tom + snare
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {118,  0, 1.0 * 6.5},
                          {-38,  0, 1.0 * 6.5}, } } }},
        // 17: Synth voice (same tuning as instr 13), same voice as 14
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {  49, 22, 1.0 * 0.3},
                          { 101, 10, 1.0 * 0.3},
                          {  85, 10, 1.0 * 0.5},
                          {  97, -2, 1.0 * 0.3},
                          {  19, -2, 1.0 * 0.3}} } }},
        // 18: Begin of synthbrass (>A = 9)
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {  63,  9, 1.0 * 0.5},
                          {  63, -3, 1.0 * 0.5}} } }},
        // 19: Short flute blow (may have woodblocks in begin), also some distoning
        //               (F + unidentified tone)
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { { 116,  5, 1.0 * 0.5},
                          {  77,  5, 1.0 * 0.5}} } }},
        // 20: Tom + bassdrum
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {118,-12, 1.0 * 2.2}, // melodic tom
                          {117,-12, 1.0 * 2.2}, // taiko
                          {-35,  0, 1.0 * 5.5} } } }},

        // 21: Chord: D#,A#,C#
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { { 28,  3-12, 1.0 * 0.6},
                          { 28,  8-12, 1.0 * 0.6},
                          { 28, 13-12, 1.0 * 0.6},
                          { 30,  3,    1.0 * 0.6},
                          { 30,  8,    1.0 * 0.6},
                          { 30, 13,    1.0 * 0.6} } } }},

        // 22: Chord: D#,G#,>C
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { { 28,  3-12, 1.0 * 0.6},
                          { 28,  8-12, 1.0 * 0.6},
                          { 28, 12-12, 1.0 * 0.6},
                          { 30,  3,    1.0 * 0.6},
                          { 30,  8,    1.0 * 0.6},
                          { 30, 12,    1.0 * 0.6} } } }},

        // 23: Chord: D# minor (D#,F#,A#)
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { { 28,  3-12, 1.0 * 0.6},
                          { 28,  6-12, 1.0 * 0.6},
                          { 28, 10-12, 1.0 * 0.6},
                          { 30,  3,    1.0 * 0.6},
                          { 30,  6,    1.0 * 0.6},
                          { 30, 10,    1.0 * 0.6} } } }},

        // 24: Void (Bisqwit improvised, sounds good); +1
        Instr_spec{ std::map<int,std::vector<Instr_spec::Part>>{ {
                     0, { {  1,    1,  0.5},
                          { 26, 1-12,  0.5}, } } }},
#endif
    };

    static const double GlobalFineTuning = 0.0; // Default
    //static const double GlobalFineTuning = 2.0; // 4-mat's That Old Magic

    struct MODchannelState
    {
        struct State
        {
            int midichannel;   // Which MIDI channel this note was triggered on

            int note;          // Which MIDI note was triggered
            int volume;        // Last volume

            double desired_note;

            struct Specs
            {
                int    ins_no;
                int    sample_offset;
                size_t index;
            } specs_id;
        };
        std::vector<State> state;
    };
    std::vector<MODchannelState> MODchannels(32);

    static const unsigned MaxMIDIchannel = 32*16; // 16 is standard

    unsigned MIDIchannelUsers[MaxMIDIchannel]  = { 0 };
    unsigned MIDIchannelVolume[MaxMIDIchannel] = { 0 };
    int      MIDIchannelPatch[MaxMIDIchannel]  = { 0 };
    int      MIDIchannelBends[MaxMIDIchannel]  = { 0 };
    int      MIDIchannelPans[MaxMIDIchannel]  = { 0 };
    int BendSense = 24;

    const unsigned PercussionChannel = 9;

    double time_total = 0;
    unsigned time_total_ms = 0;
    MIDIfile midifile;

    inline double speed_difference_to_seminotes(double diff)
    {
        // speed1 = 2 ^ (note1 / 12) ; note1 = 12 * log2(speed1)
        // speed2 = 2 ^ (note2 / 12) ; note2 = 12 * log2(speed2)
        // notedifference = 12 * (log2(speed2) - log2(speed1))
        // notedifference = 12 * log2(speed2 / speed1)
        //return 12 * std::log2(diff);
        return std::log(diff) * 17.312340490667560889;
    }

    void StateOff(unsigned chno) // Note inactive
    {
        // Keyoff all the MIDI notes active that are representing this key
        auto& c = MODchannels[chno];
        for(auto s: c.state)
        {
            //midifile[chno].KeyOff(s.midichannel, s.note, 127);
            midifile[chno].KeyOn(s.midichannel, s.note, 0); // keyoff
            MIDIchannelUsers[s.midichannel] -= 1;
        }
        c.state.clear();
    }
    void State(unsigned chno, int ins_no,
               int base_note,               // Base note. May change while playing in portamento.
               int volume,
               double sample_offset,
               double keyon_sample_speed,   // Sample speed at face value
               double current_sample_speed,
               double pan)
    {
        //if(chno != 2) return;

        auto& c = MODchannels[chno];
        /*chno   = 0;*/
        /*volume = 127;*/
        volume = volume*127/64;
        ins_no += 1;

        /*
        fprintf(stderr, "State(%u,%d, n(%d),v(%d), %f,%f,%f, %u active\n",
            chno,ins_no,base_note,volume, sample_offset,keyon_sample_speed,
            current_sample_speed,
            (unsigned) c.state.size()
        );*/


        /* Step 1: Check whether the notes currently playing are adequate.
         *         A: If they are wrong, they must be keyoff'd first.
         *         B: If they are missing, they must be keyon'd.
         * Step 2: Check whether the notes need updating.
         */
        const auto& specs = instr_specs[ins_no];

        int expected_sample_begin_offset = 0;
        std::vector<Instr_spec::Part> expected_specs;

        for(auto p: specs.formula)
            if(sample_offset >= p.first)
            {
                expected_sample_begin_offset = p.first;
                expected_specs               = p.second;
            }

        // Go through all currently active notes.
        // Remove any that are not supposed to be playing.
        c.state.erase( std::remove_if(c.state.begin(), c.state.end(),
                    [&](const MODchannelState::State& s) -> bool
                    {
                        bool ok =
                           s.specs_id.ins_no        == ins_no
                        && s.specs_id.sample_offset == expected_sample_begin_offset
                        && s.specs_id.index < expected_specs.size();
                        if(!ok)
                        {
                            //midifile[chno].KeyOff(s.midichannel, s.note, 127);
                            midifile[chno].KeyOn(s.midichannel, s.note, 0);
                            MIDIchannelUsers[s.midichannel] -= 1;
                        }
                        return !ok;
                    }), c.state.end());

        // Now go through what _should_ be playing.
        // If the note is not playing, trigger it.
        // Then ensure it has the proper straits.
        int suggested_melodic_channel = -1;
        if(0)
        {
            auto j = std::find_if(c.state.begin(), c.state.end(),
                    [&](const MODchannelState::State& s) -> bool
                    {
                        return
                           s.specs_id.ins_no        == ins_no
                        && s.specs_id.sample_offset == expected_sample_begin_offset;
                    });
            if(j != c.state.end())
                suggested_melodic_channel = j->midichannel;
        }

        for(size_t a=0; a<expected_specs.size(); ++a)
        {
            const auto& part = expected_specs[a];

            int vol = volume * part.volume;
            if(vol > 127) vol = 127; if(vol < 1) vol = 1;
            double pitch_diff = (current_sample_speed / keyon_sample_speed);
            double pitch_diff_seminotes = speed_difference_to_seminotes(pitch_diff);

            auto calc_bend = [=] (double bend_notes) -> int
            {
                int bend = 0x2000 + (bend_notes) * 0x2000 / double(BendSense);
                if(bend < 0)      bend = 0;
                if(bend > 0x3FFF) bend = 0x3FFF;
                return bend;
            };

            auto i = std::find_if(c.state.begin(), c.state.end(),
                    [&](const MODchannelState::State& s) -> bool
                    {
                        return
                           s.specs_id.ins_no        == ins_no
                        && s.specs_id.sample_offset == expected_sample_begin_offset
                        && s.specs_id.index == a;
                    });

            bool missing_keyon = false;

            if(i == c.state.end())
            {
                // Insert the note, for it was not found
                // Find which MIDI channel may be used
                int channel = PercussionChannel; // Choose percussion channel
                if(part.patch > 0) // Melodic instrument? Choose melodic channel
                {
                    if(suggested_melodic_channel >= 0
                    && part.patch == MIDIchannelPatch[suggested_melodic_channel])
                        channel = suggested_melodic_channel;
                    else
                        channel = -1;

                    if(channel == -1)
                    {
                        // Find an unused MIDI channel
                        unsigned begin = (chno*16) % MaxMIDIchannel;
                        for(unsigned a=begin; a<MaxMIDIchannel; ++a)
                            if((a%16) != PercussionChannel && MIDIchannelUsers[a] == 0)
                                { channel = a; break; }

                        // All channels are in use?
                        if(channel == -1)
                        {
                            // Find a channel that happens to have the same MIDI patch
                            for(unsigned a=begin; a<MaxMIDIchannel; ++a)
                                if(MIDIchannelPatch[a] == part.patch)
                                    { channel = a; break; }

                            // Did not find? Choose a random MIDI channel
                            // Anything but percussion though.
                            if(channel == -1)
                            {
                                channel =
                                    std::rand() % (MaxMIDIchannel*15/16);
                                channel = channel%15 + 16*(channel/15);
                                if(channel%16 >= PercussionChannel) ++channel;
                            }
                        }
                        suggested_melodic_channel = channel;
                    }
                }

                MODchannelState::State state;
                state.midichannel = channel;
                state.desired_note= base_note + part.tuning + GlobalFineTuning;
                state.note        = state.desired_note + 0.5; // This introduces rounding error
                state.volume      = MIDIchannelVolume[channel];
                state.specs_id    = { ins_no, expected_sample_begin_offset, a };

                // Choose patch, if melodic
                if(part.patch > 0)
                {
                    if(MIDIchannelPatch[channel] != part.patch)
                    {
                        midifile[chno].Patch(channel, part.patch-1);
                        MIDIchannelPatch[channel] = part.patch;
                    }
                }
                else
                {
                    // Percussion works differently
                    state.volume = vol;
                    // Record the note here, so that it can be properly disabled
                    state.note = -part.patch;
                }
                i = c.state.insert(c.state.end(), std::move(state));
                MIDIchannelUsers[channel] += 1;
                missing_keyon = true;
            }

            // We have note. Figure out which pitch we have active (without bends)
            if(1)if(part.patch > 0) // Only do pitch-bends if melodic channel
            {
                // Figure out how much we're off from current goal
                int bend = calc_bend(i->desired_note + pitch_diff_seminotes - i->note);
                if(bend != MIDIchannelBends[i->midichannel])
                {
                    midifile[chno].Wheel(i->midichannel, bend);
                    MIDIchannelBends[i->midichannel] = bend;
                }
            }

            int panctrl = 0x40 + 0x3F * pan;
            if(panctrl < 1) panctrl = 1;
            if(panctrl > 127) panctrl = 127;

            if(panctrl != MIDIchannelPans[i->midichannel])
            {
                midifile[chno].Control(i->midichannel, 10, panctrl);
                MIDIchannelPans[i->midichannel] = panctrl;
            }

            if(missing_keyon)
            {
                midifile[chno].KeyOn(i->midichannel, i->note,
                    i->midichannel == PercussionChannel ? i->volume : 127);
            }

            // Figure out if we need to adjust the volume:
            if(vol != i->volume)
            {
                i->volume = vol;
                if(part.patch > 0)
                {
                    auto& v = MIDIchannelVolume[i->midichannel];
                    if(v != i->volume)
                        midifile[chno].Control(i->midichannel, 11, v = i->volume);
                }
                else
                {
                    // For percussion, use the aftertouch
                    if(!missing_keyon)
                        midifile[chno].KeyTouch(i->midichannel, i->note, i->volume);
                }
            }
        }
    }
    void Commit()
    {
    }
    void Elapse(double seconds)
    {
        time_total += seconds;
        unsigned new_seconds = time_total*1e3;
        unsigned get_time = new_seconds - time_total_ms;
        time_total_ms += get_time;
        //fprintf(stderr, "Elapsed %g seconds, adding %g\n", seconds, get_time/1e3);
        midifile.AddDelay( get_time );
    }
    void Init()
    {
        for(unsigned a=0; a<MaxMIDIchannel; ++a)
        {
            // Set RPN 0
            midifile[0].Control(a, 100, 0);
            midifile[0].Control(a, 101, 0);
            midifile[0].Control(a, 6, BendSense); // Pitch-bender sensitivity
            int vol = 127;
            if(a == PercussionChannel) vol = 127;
            midifile[0].Control(a, 7, vol); // Set master volume
            midifile[0].Control(a, 11, MIDIchannelVolume[a] = 64); // Set expression
            midifile[0].Control(a, 10, MIDIchannelPans[a] = 0x40); // Set pan = middle
        }
    }
    void Finish()
    {
        midifile.Finish();
        FILE* fp = std::fopen("midifile.mid", "wb");
        const unsigned char* buf = &*midifile.begin();
        std::fwrite(buf, 1, midifile.size(), fp);
        std::fclose(fp);
        fprintf(stderr, "Wrote %ld bytes into midifile.mid\n", (long) midifile.size());
    }
}

static void AudioWrite(const void* Buf, unsigned nbytes)
{
    fwrite(Buf, 1, nbytes, stdout); // stdout output only
}

static const double HertzRatioPCM = (8363.0 * 1712.0 - 400.0) / SamplingRate;
//^This value is from ST3 specs (14317056).
static const double HertzRatioAdLib = (8363.0 * 1712.0 - 400.0) * (2.0/3.0);
//FIXME: ^Very slightly inexact. What is the correct value?
static double NoteHzTable[195];
static int OPL_Ports[9] = {0,1,2, 8,9,10, 16,17,18}, OPL_Pan[18];
#if HAS_YMF262
static void OPL_Byte(int o, int index, int byte)
    { YMF262Write(0,o*2,index); YMF262Write(0,1+o*2,byte); }
#else
static void OPL_Byte(int,int,int) { }
#endif
static void OPL_KeyOff(int c) { OPL_Byte(c/9,0xB0+(c%9),  0); }
static void OPL_Touch(int c, const unsigned char* D, int vol)
    { OPL_Byte(c/9,0x40+OPL_Ports[c%9], (D[2]&0xC0) | (63-((63-(D[2]&63))*vol/63)));
      OPL_Byte(c/9,0x43+OPL_Ports[c%9], (D[3]&0xC0) | (63-((63-(D[3]&63))*vol/63)));
      OPL_Byte(c/9,0xC0+(c%9), (D[10]&0xCF) | OPL_Pan[c]); }
static void OPL_NoteOff(int c)
    { OPL_KeyOff(c); OPL_Touch(c, (const unsigned char*)&c, 0); }
static void OPL_NoteOn_UpdatePitch(int c, int hertz)
    { int oct=0; for(; hertz>0x1FF; ++oct) hertz >>= 1; if(oct > 7) oct = 7;
      OPL_Byte(c/9,0xA0+(c%9), hertz & 0xFF);
      OPL_Byte(c/9,0xB0+(c%9), 0x20 | ((hertz>>8)&3) | ((oct&7)<<2)); }
static void OPL_Patch(int c, const unsigned char* D)
    { OPL_Byte(c/9,0x20+OPL_Ports[c%9], D[0]);
      OPL_Byte(c/9,0x60+OPL_Ports[c%9], D[4]);
      OPL_Byte(c/9,0x80+OPL_Ports[c%9], D[6]);
      OPL_Byte(c/9,0xE0+OPL_Ports[c%9], D[8]&3);
      OPL_Byte(c/9,0x23+OPL_Ports[c%9], D[1]);
      OPL_Byte(c/9,0x63+OPL_Ports[c%9], D[5]);
      OPL_Byte(c/9,0x83+OPL_Ports[c%9], D[7]);
      OPL_Byte(c/9,0xE3+OPL_Ports[c%9], D[9]&3);
      OPL_Byte(c/9,0xC0+(c%9), (D[10]&0xCF) | OPL_Pan[c]);
    }

template<int Radius, bool check>
static inline double Lanczos_pi(double x_pi)
{
    if(x_pi == 0.0) return 1.0;
    if(check && (x_pi <= (-Radius*M_PI)
              || x_pi >= ( Radius*M_PI))) return 0.0;
    double x_pi_div_Radius = x_pi / Radius;
    return sin(x_pi) * sin(x_pi_div_Radius) / (x_pi * x_pi_div_Radius);
}

static int Clamp(int value, int mi, int ma)
{
    return value<mi?mi : (value>ma ? ma : value);
}

namespace s3m
{
    /* The S3M file header. It is changed during playback. */
    struct s3mheader
    {
        char name[28];    // song name
        unsigned char eofchar, typ, dummy[2];
        unsigned short n_ord, n_ins, n_pat, flags, cwtv, ffi;
        char scrm[4];
        unsigned char Vxx,Axx,Txx,mastervolume,uc,dp,dummy2[8];
        unsigned short special;
        unsigned char channelsettings[32];
    } header;

    /* The S3M instrument header */
    struct s3minst /* this mostly follows the structure in the file */
    {
        unsigned char type;
        char filename[12];
        unsigned char filepos[3];
        int32_t length, loopbegin, loopend; // This doubles as D00..D0B for AdLib
        unsigned char volume, dummy, packflag, flags;//
        unsigned long c4spd;
        unsigned char* sampledata; // internal, used in runtime
        double c4spd_factor;       // internal, used in runtime
        unsigned char name[28], scri[4];
    }* inst;

    /* The rest of this struct is filled on demand */
    unsigned char orders[256], **patdata, **patend;
    unsigned n_channel;
    bool has_pcm_channels, has_adlib_channels;

    /* Runtime status of the song. Stored here because they
     * cannot be function-locals if the Callback() function
     * is to be periodically called. */
    double   VolumeNormalizer;
    int      Bxx,Cxx,SBchannel,SEx, row,order,pattern,repeat, in_playloop;
    unsigned frame, ArpeggioShift;
    const    unsigned char* patptr;

    /* This struct stores the runtime status of a channel */
    struct Channel
    {
        double sample_offset;  // Offset into the instrument's PCM data
        double period;         // Period (the inverse of pitch)
        double slideto_period; // The period for note to portamento into
        double sample_speed;   // Number of samples played per second
        double live_sample_speed; // Sample_speed affected by arpeggio
        float  pan;            // Panning position (0-1=left-right, .5=middle)

        double keyon_sample_speed; // The speed that the note was keyon'd. For MIDI.

        unsigned char base_note;   // Last note (octave * 12 + note)
        unsigned char last_ins: 7; // Last instrument
        unsigned glissando    : 1; // If nonzero, portamento is rounded to nearest semitone
        unsigned vib_offset   : 6;
        unsigned trem_offset  : 6;
        unsigned loops_remain : 4;
        
        signed char volume;          // Currently mixing volume
        signed char volume_backup;   // Backup of the volume saved by tremor opcode
        unsigned loop_begin_row    : 6;
        unsigned char funk         : 4;
        unsigned zerovolume_frames : 14;
        unsigned char arpeggio;
        unsigned char funkcounter;
        unsigned funkoffset;         // Used by FunkRepeat

        // Cache last command parameters, so e.g. D00 works
        unsigned char lastG,lastD,lastH,lastE,lastR;
        unsigned char lastF,lastJ,lastI,lastQ,lastU;

        Channel() :
            sample_offset(0), period(0), pan(0.5), base_note(255), last_ins(0),
            glissando(0), loops_remain(0), volume(0), funk(0), zerovolume_frames(0),
            arpeggio(0), funkcounter(0) { }
    }* channel;

    /* Methods */
    void Load(const char* filename)
    {
        FILE* fp = fopen(filename, "rb"); assert(fp != NULL);
        fread(header.name, 1, 0x60, fp); /* Read S3M file header */
        assert(memcmp(header.scrm, "SCRM", 4) == 0); // Verify valid module
        assert(header.n_ord <= 256); std::memset(orders, 255, sizeof(orders));
        fread(&orders, 1, header.n_ord, fp); // Read orders
        assert(header.n_ins <= 99); unsigned short insptr[99]; fread(&insptr,2,header.n_ins,fp);
        assert(header.n_pat <= 100);unsigned short patptr[100];fread(&patptr,2,header.n_pat,fp);
        unsigned char pansettings[32] = { 0 };
        if(header.dp == 252) fread(&pansettings, 1, 32, fp);

        has_pcm_channels = has_adlib_channels = false;
        n_channel = 1;
        for(unsigned c=0; c<32; ++c)
            if(header.channelsettings[c] < 16)
                { n_channel = c+1; has_pcm_channels = true; }
            else if(header.channelsettings[c] < 16+9)
                { n_channel = c+1; has_adlib_channels = true; }

        channel = new Channel[n_channel];

        for(unsigned c=0; c<n_channel; ++c)
            channel[c].pan = (pansettings[c] & 0x20)
                ? (pansettings[c] & 0x0F) / 15.0
                : ((header.mastervolume & 0x80)
                    ? (header.channelsettings[c]<8 ? 0.2 : 0.8)
                    : 0.5);

        inst = new s3minst[header.n_ins];
        for(unsigned i=0; i<header.n_ins; ++i)
        {
            s3minst& ins = inst[i];
            fseek(fp, insptr[i]*16UL, SEEK_SET);
            fread(&ins, 1, 0x50, fp);
            //assert(memcmp(ins.scri, "SCRS", 4) == 0); // Verify valid instrument
            if(!ins.c4spd) ins.c4spd = 8363;
            ins.c4spd_factor = (8363.0*16.0*1712.0) / ins.c4spd;
            //if(ins.c4spd) ins.c4spd_factor = 16*1712;
            // ^This will be explained later
            if(ins.type != 1) continue; // Only PCM samples have PCM data
            if(ins.length > 64000l) ins.length = 64000l;
            if(ins.flags & 1) /* This identifies a looping sample */
            {
                assert(ins.loopbegin < ins.length);
                assert(ins.loopend  <= ins.length);
            }
            unsigned long smppos = ins.filepos[1]*0x10UL
                                 + ins.filepos[2]*0x1000UL
                                 + ins.filepos[0]*0x100000UL;
            fseek(fp, smppos, SEEK_SET);
            ins.sampledata = new unsigned char[ins.length];
            assert(ins.sampledata != NULL);
            fread(ins.sampledata, 1, ins.length, fp);
        }

        patdata = new unsigned char* [header.n_pat];
        patend  = new unsigned char* [header.n_pat];
        for(unsigned p=0; p<header.n_pat; ++p)
        {
            unsigned short length = 2;
            fseek(fp, patptr[p] * 16UL, SEEK_SET);
            if(patptr[p]) fread(&length, 1, 2, fp);
            patdata[p] = new unsigned char [ length ];
            assert(patdata[p] != NULL);
            fread(patdata[p], 1, length, fp); // Read packed pattern data
            patend[p] = patdata[p] + length - 2;
        }
        fclose(fp);

        in_playloop = 0;
        row = order = 0;
    }

    struct Slot
    {
        unsigned channel, note, instrument, volume, command, infobyte;
        void Print()
        {
            static const char notenames[] = "C-C#D-D#E-F-F#G-G#A-A#B-12131415";
            //std::fprintf(stderr, "c%02u ", channel);
            std::fprintf(stderr, note==255?"\33[31m...":note==254?"\33[31m^^^":
                            "\33[31m%.2s%u", notenames+2*(note&15), note>>4);
            std::fprintf(stderr, instrument?"\33[34m%02u":"\33[34m..", instrument);
            std::fprintf(stderr, volume==255?"\33[32m..":"\33[32m%02u", volume);
            if(command) std::fprintf(stderr, "\33[m%c%02X", command+64, infobyte);
        }

        /* Unpack a slot from packed S3M pattern data */
        static Slot Read(const unsigned char*& p)
        {
            unsigned char byte = *p++;
            Slot result = {byte & 0x1Fu, 255,0, 255, 0,0};
            if(byte & 0x20) { result.note = *p++; result.instrument = *p++; }
            if(byte & 0x40) { result.volume = *p++; }
            if(byte & 0x80) { result.command = *p++; result.infobyte = *p++; }
            return result;
        }
    };

    static long SamplePosition(const s3minst& ins, long s)
    {
        if((ins.flags & 1) && s >= ins.loopend) /* loop? */
            s = ins.loopbegin + (s-ins.loopbegin) % (ins.loopend-ins.loopbegin);
        return s;
    }

    /* Load a PCM sample from an instrument */
    static double LoadSample(const s3minst& ins, double center, double inc_rate)
    {
        if(DoLanczosFiltering)
        {
            const int Radius  = 2;
            const double blur = 1;
            const double factor = inc_rate;
            double scale  = factor; if(scale < 1.0) scale = 1.0;
            double support = scale * (Radius * blur);
            if(support < 0.5) { support = 0.5; scale = 1.0; }

            const long start = (long)(center - support + 0.5);
            const long stop  = (long)(center + support + 0.5);
            const long nmax = stop - start;

            const double pi_scale = M_PI / scale;
            double s        = (start - center + 0.5) * pi_scale;
            double result   = 0.0, density = 0.0;
            for(long n=0; n<nmax; ++n, s += pi_scale)
            {
                const double contribution = Lanczos_pi<Radius,true> (s);
                density += contribution;

                long pos = SamplePosition(ins, start + n);
                if(pos < 0) continue;
                if(pos >= ins.length) break;
                result += contribution * (ins.sampledata[pos] - 128);
            }
            if(density != 0.0) return result / density;
            return result;
        }
        long pos = SamplePosition(ins, (long) center);
        if(pos >= ins.length) return 0; // End of note
        return (int)ins.sampledata[pos] - 128;
    }

    void SetChannelSpeed(unsigned chno, double live_period)
    {
        Channel& ch = channel[chno];
        if(ch.base_note != 255)
        {
            if(live_period != 0.0)
                switch(inst[ch.last_ins].type)
                {
                    case 1: // PCM
                        ch.sample_speed = HertzRatioPCM / live_period;
                        break;
                    case 2: // Adlib melodic
                        ch.sample_speed = HertzRatioAdLib / live_period;
                        OPL_Pan[chno] = (ch.pan < 0.8)*0x10
                                      + (ch.pan > 0.2)*0x20;
                        OPL_NoteOn_UpdatePitch(chno, (int) ch.sample_speed);
                        OPL_Touch(chno,
                            (const unsigned char*)&inst[ch.last_ins].length,
                            ch.volume * header.Vxx * 63 / (64*64));
                }
            ch.live_sample_speed = ch.sample_speed;
            if(ch.arpeggio)
                ch.live_sample_speed *= NoteHzTable[(ch.arpeggio >> ArpeggioShift) & 15];
    }   }

    void InterpretRow()
    {
        for(const unsigned char* rowptr = patptr; *rowptr != 0; )
        {
            /* Parse the current slot */
            Slot slot = Slot::Read(rowptr);
            if(slot.channel >= n_channel) continue;

            if(frame==0 && repeat==0)
                { std::fprintf(stderr, "\r\33[%uC", 10+slot.channel*10);
                  slot.Print(); }

            Channel& ch = channel[slot.channel];

            unsigned noteon_frame=0, notecut_frame=999, sample_offset=0;
            unsigned volumeslide=0, pitchslide=0, portamento=0;
            unsigned tremor=0, retrig=0, vibrato=0, tremolo=0;

            #define RawOp(option,saveto,pref,suff) \
                case option: pref; saveto = slot.infobyte; suff; break;
            #define NormalOp(option,saveto,cache,suff) RawOp(option,saveto,\
                if(!slot.infobyte)slot.infobyte=ch.cache; \
                else ch.cache = slot.infobyte, suff)
            #define HalvesOp(option,saveto,cache,suff) RawOp(option,saveto,\
                if(!(slot.infobyte & 15)) slot.infobyte |= ch.cache&15; \
                if(!(slot.infobyte >> 4)) slot.infobyte |= ch.cache&0xF0; \
                ch.cache = slot.infobyte, suff)

            switch(slot.command+64) /* Parse the command */
            {
                RawOp('A', header.Axx,, )  // Axx: Set speed xx
                RawOp('T', header.Txx,, )  // Txx: Set tempo xx
                RawOp('V', header.Vxx,, )  // Vxx: Set global volume xx
                RawOp('B', Bxx,, )         // Bxx: Break to order xx row 0
                RawOp('C', Cxx,, Cxx-=Cxx/16*6)// Cxx: Cut to next pattern row xx
                RawOp('O', sample_offset,,
                    sample_offset*=256)// Oxx: Set sampledata offset
                RawOp('X', ch.pan,, ch.pan /= 255.0)          // Xxx: Set channel pan position
                NormalOp('G', portamento, lastG, )  // Gxx: Tone portamento
                HalvesOp('R', tremolo, lastR, )     // Rxy: Tremolo (volume vibrato)
                HalvesOp('H', vibrato, lastH, )     // Hxy: Pitch vibrato (times 4)
                HalvesOp('U', vibrato, lastU, vibrato |= 0x100) // Uxy: Pitch vibrato (fine)
                NormalOp('E', pitchslide, lastE, )  // Exy: Pitch slides down
                NormalOp('F', pitchslide, lastF, pitchslide |= 0x100) // Fxy: Pitch slides up
                NormalOp('D', volumeslide, lastD, ) // Dxy: Volume slides
                NormalOp('K', volumeslide, lastD, vibrato  = ch.lastH) // Kxy: H00 + Dxy
                NormalOp('L', volumeslide, lastD, portamento=ch.lastG) // Lxy: G00 + Dxy
                NormalOp('J', ch.arpeggio, lastJ, ) // Jxy: Arpeggio 50 Hz { n,n+x,n+y }
                NormalOp('I', tremor, lastI, )      // Ixy: Tremor with ontime=x, offtime=y
                NormalOp('Q', retrig, lastQ, )      // Qxy: Retrig at y frame intervals, fade by x
                case 'S':
                {   unsigned char stype = slot.infobyte>>4; slot.infobyte &= 0x0F;
                    switch(stype)
                    {
                        case 0xB: // SBx: Pattern loop
                            if(frame != 0) break;
                          { int cur=ch.loops_remain;
                            if(!slot.infobyte)
                              ch.loop_begin_row = row;
                            else if((ch.loops_remain=(cur ? cur-1 : slot.infobyte)))
                              SBchannel = slot.channel;
                            else
                              ch.loop_begin_row = row+1;
                          } break;
                        RawOp(0xC, notecut_frame,, ) // SCx: Note cut @ frame x
                        RawOp(0xD, noteon_frame,, )  // SDx: Note delay @ frame x
                        RawOp(0xE, SEx,, )           // SEx: Row repeat x times
                        RawOp(0x8, ch.pan,, ch.pan /= 15.0 )
                            // S8x: Set channel pan position
                            // ^Protracker: KarplusStrong filter
                        RawOp(0x1, ch.glissando,, )  // S1x: Set glissando control
                        RawOp(0xF, ch.funk,, )       // SFx: Set funk-repeat
                        // TODO: S2x (Set Amiga Finetune)
                        // TODO: S3x (Set vibrato waveform to x)
                        // TODO: S4x (Set tremolo waveform to x)
            }   }   }

            // Decide upon the manners of triggering the note

            if((retrig & 15) && (frame % (retrig & 15) == 0))
            {
                static const signed char add_mul[32] =
                {0,-1,-2,-4,-8,-16,0,0, 0,1,2,4,8,16,0, 0,
                 6, 6, 6, 6, 6,  6,4,3, 6,6,6,6,6, 6,9,12};
                noteon_frame = frame;
                if(frame>0 || (slot.note==255 && slot.instrument==0))
                    slot.instrument = ch.last_ins+1;
                if(frame>0)
                    slot.volume = Clamp((ch.volume + add_mul[retrig>>4])
                                        * add_mul[(retrig>>4)+16] / 6, 0,64);
            }

            if(frame == noteon_frame && (repeat == 0 || frame > 0))
            {
                if(slot.note == 254)
                    notecut_frame = frame; // ^^^ = cut the note
                else if(slot.note != 255 || slot.instrument != 0)
                {
                    /* Trigger the note */
                    if(slot.note != 255)
                        ch.base_note = (slot.note>>4) * 12 + (slot.note&0x0F);
                    if(slot.instrument != 0)
                    {
                        ch.last_ins   = slot.instrument - 1;
                        ch.funkoffset = 0;
                        if(slot.volume == 255)
                            ch.volume = inst[ch.last_ins].volume;
                        if(inst[ch.last_ins].type == 2)
                            OPL_Patch(slot.channel,
                                      (const unsigned char*)&inst[ch.last_ins].length);
                    }

                    //  period = 2 ^ (note/-12) * 8363*16*1712 / c4spd
                    ch.slideto_period =
                        inst[ch.last_ins].c4spd_factor
                        / NoteHzTable[ch.base_note];

                    if(!portamento) // retrig the note
                    {
                        if(inst[ch.last_ins].type == 2)
                        {
                            OPL_KeyOff(slot.channel);
                        }

                        MIDItracer::StateOff(slot.channel);

                        ch.sample_offset = sample_offset;
                        ch.period        = ch.slideto_period;

                        double ratio = HertzRatioPCM;
                        if(inst[ch.last_ins].type == 2) ratio = HertzRatioAdLib;

                        ch.keyon_sample_speed = ratio / ch.period;
                    }
                }
                if(slot.volume != 255)
                    ch.volume = slot.volume;
            }
            if(frame == notecut_frame)
            {
                if(inst[ch.last_ins].type == 2)
                {
                    OPL_NoteOff(slot.channel);
                    MIDItracer::StateOff(slot.channel); // TODO: Cut
                }
                ch.base_note = 255; // Cut the note */
            }

            // Apply volume changes (which always stick)
            if(ch.base_note == 255) continue; // Skip a channel made inactive

            if(volumeslide)
            {
                int n=0, fine=1, a=frame==noteon_frame, s=volumeslide;
                /**/ if((s&0xF0)==0x00) fine=0,n=-(s&0x0F);// D0y: slide down
                else if((s&0x0F)==0x00) fine=0,n= (s>>4);  // Dx0: slide up
                else if((s&0x0F)==0x0F) fine=1,n= (s>>4);  // DxF: fine slide up
                else if((s&0xF0)==0xF0) fine=1,n=-(s&0x0F);// DFy: fine slide down
                //     AB: 00 01 10 11   Truth table
                // FINE=0: T  T  F  T      A = first frame ?
                // FINE=1: F  F  T  T      B = fastmode ?
                if(fine ? a : (!a || header.cwtv==0x1300 || (header.flags&0x40)))
                    ch.volume = Clamp(ch.volume + n, 0,64);
            }
            else if(tremolo) // Apply tremolo
            {
                ch.volume = Clamp(ch.volume
                    +  std::sin(ch.trem_offset * (3.14159265*2/64))
                    * (1+(tremolo & 0x0F)), 0, 64);
                // FIXME: Tremolo magnitude probably incorrect
                if(frame > 0) ch.trem_offset += tremolo >> 4;
            }
            else if(tremor)
            {
                unsigned ontime = 1+(tremor>>4), offtime = 1+(tremor&15);
                if(frame == 0) ch.volume_backup = ch.volume;
                else if(frame%(ontime+offtime)==0) ch.volume=ch.volume_backup;
                else if(frame%(ontime+offtime)==ontime) ch.volume=0;
            }

            // Apply pitch changes that stick

            if(portamento) /* Pitch slide towards the desired note */
            {
                if(ch.period < ch.slideto_period)
                  { if((ch.period += portamento*4) >= ch.slideto_period)
                      ch.period = ch.slideto_period; }
                else if(ch.period > ch.slideto_period)
                  { if((ch.period -= portamento*4) <= ch.slideto_period)
                      ch.period = ch.slideto_period; }
            }
            else if(pitchslide) /* Pitch slide by specified amount */
            {
                int magnitude = pitchslide & 0xFF, fine = 0;
                if(magnitude >= 0xF0) { fine=2; magnitude &= 15; }
                if(magnitude >= 0xE0) { fine=1; magnitude &= 15; }
                if(fine < 2) magnitude *= 4;
                if(!fine || frame == noteon_frame)
                {
                    ch.period += (pitchslide&0x100)
                        ?-magnitude:magnitude; // down or up
                    if(ch.period < 4) ch.period = 4;
                }
            }

            double live_period = ch.period;

            // Other

            if(frame > 0 && ch.funk) /* FunkRepeat */
            {   static const unsigned char funktable[16]=
                {0,5,6,7,8,10,11,13,16,19,22,26,32,43,64,128};
                if((ch.funkcounter += funktable[ch.funk]) >= 128)
                {   ch.funkcounter = 0;
                    const s3minst& ins = inst[ch.last_ins];
                    if(ins.type == 1 && ins.loopend > ins.loopbegin)
                        ins.sampledata[ins.loopbegin
                          + (++ch.funkoffset) % (ins.loopend - ins.loopbegin)] ^= 255;
            }   }
            
            /* Amiga Filter (S0x in Protracker) appears to only
             * control the Ledlight, at least in some versions of Amiga...
             */
            
            /* KarplusStrong filter (S8x in Protracker 1.2A and 3.00B, not in 4):
             * Executed once (at frame 0) only on the row it occurs
             *
             *   A0 = loopstart
             *   A1 = A0
             *   D0 = replen*2 - 2
             * loop:
             *   [A0] = ( [A0] + [A0+1] ) / 2
             *   ++A0
             *   dbra D0, loop
             *   [A0] = ([A0] + [A1]) / 2
             */

            // Apply pitch changes that don't stick

            if(vibrato) // Apply vibrato
            {
                live_period += std::sin(ch.vib_offset * (3.14159265*2/64))
                    * ((vibrato & 0x0F) << (vibrato<0x100 ? 3 : 1));
                if(frame > 0) ch.vib_offset += (vibrato&0xF0) >> 4;
            }
            else if(portamento && ch.glissando)
            {
                // Round the pitch into the nearest semitone
                // Formula (c4spdfactor is "8363*16*1712/c4spd"):
                //   semitone = round(log2(period/c4spdfactor) * -12)
                //   newperiod = c4spdfactor * 2 ^ (semitone / -12)
                live_period /= inst[ch.last_ins].c4spd_factor;
                int semitone = 0.5 - std::log(live_period) * 17.31234049066756;
                live_period = inst[ch.last_ins].c4spd_factor;
                if(semitone >= 195) semitone = 195-1;
                live_period /= NoteHzTable[semitone];
            }

            SetChannelSpeed(slot.channel, live_period);
        }
        if(frame==0 && repeat==0) std::fprintf(stderr, "\n");
    }

    /* Play the song. This function is such designed that it can
     * be invoked from a periodic interrupt handler (timer IRQ) */
    void PlayCallback()
    {
        if(in_playloop) goto PlayLoop;
        Bxx = Cxx = -1; /* State 0: Begin song playing */

        if(0)memcpy(
            orders,
            "\x08\x04\x0C\x04\x08\x04\x0C\x04\x08\x04\x0C\x04\x08\x04\x0C\x04",
            16);
        if(0) { Bxx = 0; header.Axx = 6; }
        
        for(;;)
        {
            if(Bxx >= 0) { order = Bxx; Bxx = -1; }

            /* Load the pattern pointer */
            if(order >= header.n_ord) { order=0; continue; }
            pattern = orders[order++];
            if(pattern == 255) { order=0; continue; }
            if(pattern == 254) { continue; }

            patptr = patdata[pattern];
            assert(patptr != NULL);

            /* Skip to the desired row */
            for(row=0; row<Cxx; ++row, ++patptr)
                while(*patptr != 0) Slot::Read(patptr);
            if(row == 0)
                {for(unsigned c=0; c<n_channel; ++c)
                    channel[c].loop_begin_row = 0;}

            /* TODO: Verify whether we should reset loop_begin_row
             * when transition to new pattern */
            /* Play the pattern until its end */
            Cxx       = -1;
            SBchannel = -1;
            for(; patptr < patend[pattern]; ++patptr, ++row)
            {
                std::fprintf(stderr,
                    "\33[mO\33[34m%02X"
                    "\33[mP\33[34m%03u"
                    "\33[mR\33[34m%02u\33[m",
                    order-1, pattern, row);

                for(unsigned c=0; c<n_channel; ++c)
                {
                    channel[c].arpeggio = 0;
                    SetChannelSpeed(c, channel[c].period);
                }

                /* Parse the current row (repeat Axx frames, SEx times) */
                SEx = 0;
                for(repeat=0; repeat <= SEx; ++repeat)
                    for(frame=0; frame < header.Axx; ++frame)
                    {
                        InterpretRow();

                        /* Generate audio for this frame */
                        if(header.flags & 8) // 0vol optimizations?
                        {
                            // Shut down notes that have been silent for > 2 rows
                            for(unsigned c=0; c<n_channel; ++c)
                                if(channel[c].base_note == 255) {}
                                else if(channel[c].volume > 0)
                                    channel[c].zerovolume_frames = 0;
                                else if(++channel[c].zerovolume_frames > header.Axx*2)
                                    { channel[c].base_note=255;
                                      channel[c].zerovolume_frames=0; }
                        }

                        VolumeNormalizer = (header.mastervolume & 127) * header.Vxx / 4096.0;

                        static int IRQCounter = 0;
                        for(IRQCounter += 5 * IRQRate; IRQCounter >= header.Txx*2; IRQCounter -= header.Txx*2)
                        {
                            // Frame timing:    Each frame is 2.5 / Txx seconds long. Row = Axx frames.
                            // Arpeggio timing: Pitch changes 50 times in a second.
                            // IRQ timing:      Each IRQ cycle lasts for 1/IRQRate seconds.
                            // Sample timing:   Each PCM sample lasts for 1/SamplingRate seconds.
                            static unsigned ArpeggioCounter = 0; ArpeggioCounter += 50;
                           {unsigned n_arpeggio_shifts = ArpeggioCounter / IRQRate; ArpeggioCounter %= IRQRate;
                            if(n_arpeggio_shifts)
                            {   ArpeggioShift = (((ArpeggioShift>>2)+n_arpeggio_shifts)%3)<<2;
                                for(unsigned c=0; c<n_channel; ++c)
                                {
                                    SetChannelSpeed(c, 0.0);
                                    if(channel[c].base_note != 255
                                    && channel[c].arpeggio
                                    && inst[channel[c].last_ins].type == 2) // Non-idle AdLib channel having arpeggio
                                    {
                                        OPL_NoteOn_UpdatePitch(c, (int) channel[c].live_sample_speed);
                           }}   }   }

                           {static unsigned PCMCounter = 0; PCMCounter += SamplingRate;
                            unsigned n_pcm_samples = PCMCounter / IRQRate; PCMCounter %= IRQRate;

                            static short MixBuffer[MaxPCMsamplesPerIRQ*2] = {0};
                            #if HAS_YMF262
                            YMF262UpdateOne(0, MixBuffer, n_pcm_samples);
                            #endif

                            for(unsigned c=0; c<n_channel; ++c)
                            {
                                Channel& ch = channel[c];
                                if(ch.base_note == 255)
                                    MIDItracer::StateOff(c);
                                else
                                    MIDItracer::State(c,
                                        ch.last_ins,
                                        ch.base_note,
                                        ch.volume,
                                        ch.sample_offset,
                                        ch.keyon_sample_speed,
                                        ch.live_sample_speed,
                                        ch.pan
                                    );
                            }
                            MIDItracer::Commit();
                            MIDItracer::Elapse( double(n_pcm_samples) / double(SamplingRate) );

                            if(has_pcm_channels)
                                for(unsigned m=n_pcm_samples*2, n=0; n<m; n+=2)
                                {
                                    double result[2] = { 0.0, 0.0 };
                                    for(unsigned c=0; c<n_channel; ++c)
                                    {
                                        Channel& ch = channel[c];
                                        if(ch.base_note == 255) continue; // Idle channel
                                        const s3minst& ins = inst[ch.last_ins];
                                        if(ins.type != 1) continue; // PCM only
                                        if(!(ins.flags & 1) && ch.sample_offset >= ins.length)
                                            { ch.base_note = 255; continue; }

                                        // At each mixing step, sample reading position
                                        // is incremented by Hertz / PlaybackRate. The
                                        // value for Hertz is specified in the ST3 docs
                                        // as:     14317056 / Period,
                                        // where Period for note N =
                                        //         8363*16*1712 / (c4spd * 2^N/12))
                                        // C4spd is the middle-C finetuning value
                                        // for the instrument.
                                        // Note  0=C0, 1=C#0, 2=D0...
                                        //      12=C1...
                                        //      22=A#1, 23=B1, 24=C2...
                                        // And so on. 12 notes per octave.
                                        // The Period is altered by pitch slides and vibratos.
                                        if(ch.volume)
                                        {
                                            double smp = ch.volume * LoadSample(ins, ch.sample_offset, ch.live_sample_speed);
                                            result[1] +=      ch.pan  * smp;
                                            result[0] += (1.0-ch.pan) * smp;
                                        }
                                        ch.sample_offset += ch.live_sample_speed;
                                    }
                                    MixBuffer[n+0] = Clamp(result[0]*VolumeNormalizer+MixBuffer[n+0], -32768,32767);
                                    MixBuffer[n+1] = Clamp(result[1]*VolumeNormalizer+MixBuffer[n+1], -32768,32767);
                                }
                            AudioWrite(MixBuffer, n_pcm_samples*2*2);
                            in_playloop = 1; return;}
                        PlayLoop:;
                        } /* IRQ cycle loop */
                    } /* next frame */

                /* Goto next row */
                if(SBchannel >= 0) /* A channel triggered a pattern loop? */
                    Bxx = order-1, // Repeat the same pattern
                    Cxx = channel[SBchannel].loop_begin_row;
                if(Bxx >= 0 || Cxx >= 0) break; /* Pattern break? */
                /* This row is done, goto next */
                while(*patptr != 0) Slot::Read(patptr);
    }   }   }
};

static volatile bool done = false;
static void End(int)
{
    done = true;
}

int main(int argc, char**argv )
{
    assert( sizeof(int32_t) == 4 );
    assert( sizeof(double) == 8 );

    for(unsigned a=0; a<195; ++a) /* Table of 2 ^ (x/12) */
         NoteHzTable[a] = std::exp(a * .0577622650466621);
    /* 195 = 15*12+15, which is the maximum number of notes
     * (actually 15*12+11,  but includes bad data safety).
     * pow(x,y) = exp(log(x) * y)        This is a mathematical fact.
     * pow(x,-y) = 1 / pow(x,y)          This too. (when x > 0)
     * pow(2, a/-12) = exp(log(2)*a/-12).
     * The magic number above therefore is log(2)/-12.
     */
  #if HAS_YMF262
    YMF262Init(1, 14400000l, SamplingRate);
    OPL_Byte(0,1,0x20); OPL_Byte(0,0xBD,0); OPL_Byte(0,4,0); OPL_Byte(1,5,1);
    for(unsigned c=0; c<18; ++c) { OPL_Pan[c] = 0x30; OPL_NoteOff(c); }
  #endif
    s3m::Load(argv[1]);

    MIDItracer::Init();

    signal(SIGINT, End);

    for(;!done;) s3m::PlayCallback();

    MIDItracer::Finish();

    return 0;
}
